#!/usr/bin/python

# file:        threaded_enfore.py
# author:      ben servoz
# description: program to enforce google permissions
# notes:       partial instructions at http://ben.akrin.com/?p=3923

import base64
import json
import math
import os.path
import pprint
import pycurl
import random
import signal
import smtplib
import socket
from StringIO import StringIO
import sys
from sys import argv
import threading
import time
import urllib

# APIs needed, admin sdk, drive

# configuration
crawler_threads_count = 1
processor_threads_count = 2
domain_delegated_account_json_file = "/path/to/your/domain/account.json"
error_contact = "error_email@domain.com"
google_apps_domain = "apps.domain.com"
# todo check that there aren't more crawlers than processors, processors should always be ahead of the game and they do a lot more API calls.

# boolean used to terminate threads if SIGINT was received
sigint_received = False ;
# another boolean in case we need to bail on error
fatal_error_reached = False

# global variables
crawler_file_stack = []
crawler_file_stack_fresh = True
crawler_file_stack_lock = threading.Lock()
crawler_threads_status = []
file_ids_to_process = []
file_ids_to_process_lock = threading.Lock()
file_ids_being_processed = []
file_ids_being_processed_lock = threading.Lock()
file_ids_cooling_off = {}
file_ids_cooling_off_lock = threading.Lock()
processor_threads_status = []
all_folder_ids = []

script, root_folder_id, root_owner = argv # TODO at some point it would be nice if root_owner was automatically retrieved
# TODO sanity checking on arguments


def signal_handler( signal, frame ):
	global sigint_received
	sys.stdout.write( "SIGINT received, cleaning up threads and bailing." ) ;
	sigint_received = True ;
signal.signal(signal.SIGINT, signal_handler)


def mail( to, subject, body ):
	sender = "google-enforcer@" + socket.gethostname()
	to_list = [to]
	message = "From: <" + sender + ">\r\nTo: <" + to + ">\r\nSubject: " + subject + "\r\n\r\n" + body + "\r\n"

	try:
		smtpObj = smtplib.SMTP('localhost')
		smtpObj.sendmail( sender, to_list, message )
	except SMTPException:
		sys.stderr.write( "\n\n==========\nERROR: unable to send email to report erroneous condition!\n==========\n\n" )


def compare_permissions( reference_permissions, current_permissions ):
	if len(reference_permissions)!=len(current_permissions):
		return False
	else:
		for reference_permission in reference_permissions:
			found_it = False
			for current_permission in current_permissions:
				if compare_permission( reference_permission, current_permission ):
					found_it = True
					break
			if not found_it:
				return False
				break # not like we'll get there but hey.
	# we made it! they must be the same
	return True


def compare_permission( reference_permission, current_permission ):
	if reference_permission['type']=="anyone" and current_permission['type']=="anyone":
		return True
	#elif reference_permission['type']=="anyone" or current_permission['type']=="anyone":
	#	# nothing to do, clearly not the same but we want to make sure the next elif doesn't take place since the 'name' and 'domain' keys won't exist
	#	pass
	elif reference_permission['emailAddress']==current_permission['emailAddress'] and reference_permission['domain']==current_permission['domain'] and reference_permission['role']==current_permission['role'] and reference_permission['type']==current_permission['type']:
		# we don't compare ['name'] here because not all permissions have it
		return True
	return False


def get_owner_permission( permissions ):
	for permission in permissions:
		if permission['role']=="owner":
			# as far as I know, there can only ever be 1 owner; Google may change that, their current data structure would certainly allow it
			return permission
	return None


def processor( id, gd, reference_permissions ):
	global file_ids_to_process, file_ids_to_process_lock, processor_threads_status, fatal_error_reached, file_ids_cooling_off, file_ids_cooling_off_lock


	sys.stdout.write( "* processor " + str(id) + " starting\n" )
	while not sigint_received and not fatal_error_reached:

		if id==0: # first processor gets to report on queue size
			sys.stdout.write( "-=| current queue size: " + str(len(file_ids_to_process)) + " |=-\n" )

		file_ids_being_processed_lock.acquire()
		file_ids_to_process_lock.acquire()
		if len(file_ids_to_process)>0:
			processor_threads_status[id] = True
			random_index = random.randrange( 0,len(file_ids_to_process) )
			file_id = file_ids_to_process.pop( random_index )
			file_ids_being_processed.append( file_id )
			file_ids_to_process_lock.release()
			file_ids_being_processed_lock.release()

			# the meat of what we do
			sys.stdout.write( "* processor " + str(id) + " / " + file_id + "\n" )

			keep_going = True
			data = gd.get_file( file_id )
			if data==False:
				# Either the file went away before we had a chance to do something with it; or we don't have permission as the root owner to view that file (if it was moved from another context for example).
				# there is nothing more we can do, maybe it really went away; or we count on the other context to repatriate the file and move it into its "Orphaned Files", because as far as this ohter context is concerned, the file doesn't have a legit parent ID (the parent ID maps to our current context and permissions don't cross contexts).
				# it does mean that one cannot move files between contexts, one way to fix this would be to have an overarching root type account which has access to everything but there are performance and security reasons why we don't want to do that.
				keep_going = False
				sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
				mail( error_contact, "unable to get_file", root_owner + " - unable to get_file with ID " + file_id )

			if keep_going and gd.is_file_trashed( file_id, data ):
				sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> file is trashed\n" )
				# TODO, I don't think we should have trashed files as the owner service account, at this point maybe we want to untrash it and put it in the "Orphaned Files" directory
				keep_going = False

			permissions = False
			owner_has_changed = False
			if keep_going:
				permissions = gd.list_permissions( file_id )
				if permissions==False:
					keep_going = False
					sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
					mail( error_contact, "unable to list_permissions", root_owner + " - unable to list_permissions for ID " + file_id )

			if keep_going:
				if compare_permissions( reference_permissions, permissions ):
					sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> permissions ok\n" )
				else:
					sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> permissions erroneous, fixing\n" )

					# first we verify/update ownsership
					#   we find the reference owner
					reference_owner_permission = get_owner_permission( reference_permissions )
					if reference_owner_permission==None:
						sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> can't find owner for reference permission\n" )
						mail( error_contact, "unable to find reference owner permission", root_owner + " - unable to find reference owner permission. we were working on ID " + file_id + ", not that this should be relevant" )
						keep_going = False
						fatal_error_reached = True
					#   then we find the current owner
					owner_permission = get_owner_permission( permissions )
					if owner_permission==None:
						sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> can't find owner for permission\n" )
						mail( error_contact, "unable to find owner permission", root_owner + " - unable to find owner permission for ID " + file_id )
						keep_going = False
					#   we can only do further work on the file if the current owner is in our domain
					if keep_going and owner_permission['domain']!=google_apps_domain:
						sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> owner is not in domain I can sub-as. I can't do anything about it.\n" )
						keep_going = False
					if keep_going and not compare_permission( reference_owner_permission, owner_permission ):
						# looks like we need to update the owner, then we'll do more work as needed
						sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> changing owner from " + owner_permission['emailAddress'] + " to " + reference_owner_permission['emailAddress'] + "\n" )
						# before we actually change anything with the file, we add it to the cool-off queue so the watcher doesn't re-add it because of our own changes
						file_ids_cooling_off_lock.acquire()
						file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
						file_ids_cooling_off_lock.release()
						result = gd.change_ownership( file_id, owner_permission, reference_owner_permission ) # the reason we pass the reference owner permission is because we'll be subing as it
						# the API call may take a while so we reset the expiration after it
						file_ids_cooling_off_lock.acquire()
						file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
						file_ids_cooling_off_lock.release()
						#if result==404: # the file went away before we had a chance to run this operation, we forget about it
						#	keep_going = False
						#else:
						# let's get a refreshed set of permissions to check what we have just done and for subsequent operations
						permissions = gd.list_permissions( file_id )
						if permissions==False:
							keep_going = False
							sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
							mail( error_contact, "unable to list_permissions", root_owner + " - unable to list_permissions for ID " + file_id )
						if keep_going:
							owner_permission = get_owner_permission( permissions )
							if not compare_permission( reference_owner_permission, owner_permission ):
								sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> owner change doesn't seem to have taken effect.\n" )
								mail( error_contact, "owner change didn't take effect", root_owner + " - I tried to change ID " + file_id + "'s owner to " + reference_owner_permission['emailAddress'] + ", but it didn't take" )
								keep_going = False
						if keep_going:
							# the owner has changed, when we're done reviewing other permissions we'll want to grab a fresh copy of the file
							owner_has_changed = True
					if keep_going:
						# ok at this point, the permissions still need to be worked on but at least the owner is correct
						# adding permissions that need to be added
						for reference_permission in reference_permissions:
							if reference_permission['type']!="owner": # owner is right by this point
								found_it = False
								for permission in permissions:
									if compare_permission( reference_permission, permission ):
										found_it = True
										break
								if not found_it:
									sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> adding permission " + reference_permission['emailAddress'] + "/" + reference_permission['role'] +  "\n" )
									file_ids_cooling_off_lock.acquire()
									file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
									file_ids_cooling_off_lock.release()
									gd.insert_permission( file_id, reference_permission )
									file_ids_cooling_off_lock.acquire()
									file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
									file_ids_cooling_off_lock.release()
									# todo error checking after API call?
						# removing permissions that need to be removed
						for permission in permissions:
							if permission['type']!="owner": # owner is right by this point
								found_it = False
								for reference_permission in reference_permissions:
									if compare_permission( reference_permission, permission ):
										found_it = True
										break
								if not found_it:
									sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> removing permission " + permission['emailAddress'] + "/" + permission['role'] +  "\n" )
									file_ids_cooling_off_lock.acquire()
									file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
									file_ids_cooling_off_lock.release()
									gd.delete_permission( file_id, permission )
									file_ids_cooling_off_lock.acquire()
									file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
									file_ids_cooling_off_lock.release()
									# todo error checking after API call?
					if keep_going:
						# let's get a refreshed batch of permissions
						permissions = gd.list_permissions( file_id )
						if permissions==False:
							keep_going = False
							sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
							mail( error_contact, "unable to list_permissions", root_owner + " - unable to list_permissions for ID " + file_id )
					if keep_going:
						# at this point the permissions better match with the reference permissions
						if not compare_permissions( reference_permissions, permissions ):
							keep_going = False
							sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> looks like the permissions still don't match at the end of the permission syncing process...\n" )
							mail( error_contact, "permissions don't match at the end of the permission syncing process", root_owner + ", ID: " + file_id )

			# ok we're done with the permissions, phew. now if the owner has changed, we "get" a new version of the file because now that we are the owner, we could be seing things differently
			data = gd.get_file( file_id )
			if data==False:
				# Either the file went away before we had a chance to do something with it; or we don't have permission as the root owner to view that file (if it was moved from another context for example).
				# there is nothing more we can do, maybe it really went away; or we count on the other context to repatriate the file and move it into its "Orphaned Files", because as far as this ohter context is concerned, the file doesn't have a legit parent ID (the parent ID maps to our current context and permissions don't cross contexts).
				# it does mean that one cannot move files between contexts, one way to fix this would be to have an overarching root type account which has access to everything but there are performance and security reasons why we don't want to do that.
				keep_going = False
				sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
				mail( error_contact, "unable to get_file", root_owner + " - unable to get_file with ID " + file_id )

			# noping writercanshare
			if keep_going:
				if gd.is_file_writers_can_share( file_id, data ):
					sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> setting writersCanShare to false\n" )
					file_ids_cooling_off_lock.acquire()
					file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
					file_ids_cooling_off_lock.release()
					gd.set_writers_can_share_to_false( file_id )
					file_ids_cooling_off_lock.acquire()
					file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
					file_ids_cooling_off_lock.release()
					# TODO error checking after API call

			if keep_going:
				orphaned_file = True
				for parent in data['parents']:
					if parent['isRoot']==False:
						orphaned_file = False
						break
				if orphaned_file:
					sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> file is orphaned from my perspective, let's try to find the parent\n" )
					found_a_parent = False
					permissions = gd.list_permissions( file_id )
					for permission in permissions:
						if permission['type']=="group":
							group_members = gd.list_group_members( permission['emailAddress'] )
							for group_member in group_members:
								if group_member['type']=="USER" and group_member['role']=="MEMBER":
									email = group_member['email']
									domain = email.split( "@" )[1]
									if domain==google_apps_domain:
										sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> " + group_member['email'] + ", do you know parent?\n" )
										# TODO, make sure that we find someone who knows the parent
										result = gd.get_file( file_id, email )

										orphaned_file = True
										parent_id = ""
										for parent in result['parents']:
											if parent['isRoot']==False:
												orphaned_file = False
												sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> looks like you do :)\n" )
												parent_id = parent['id']
												break

										if not orphaned_file:
											# yay, we found someone with access to the parent, let's see how owns the parent and give up ownership of file_id to the account
											result = gd.get_file( parent_id, email )
											if result==False:
												sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> looking at parent id " + parent_id + " either the file went away before I had a chance to process it, or I don't have permission to do anything with it\n" )
												mail( error_contact, "unable to list_permissions", root_owner + " - unable to get_file for ID " + parent_id )
												keep_going = False
											if keep_going:
												new_owner = result['owners'][0]['emailAddress']
												sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> removing any residual permissions, to avoid conflict with new owner picking up the file and fixing it\n" )
												for permission2 in permissions:
													if permission2['role']!="owner":
														file_ids_cooling_off_lock.acquire()
														file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
														file_ids_cooling_off_lock.release()
														gd.delete_permission( file_id, permission2 )
														file_ids_cooling_off_lock.acquire()
														file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
														file_ids_cooling_off_lock.release()
												sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> giving up ownership, changing owner from " + root_owner + " to " + new_owner + "\n" )
												file_ids_cooling_off_lock.acquire()
												file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
												file_ids_cooling_off_lock.release()
												gd.change_ownership( file_id, {'emailAddress':root_owner}, {'role':'owner','type':'user','emailAddress':new_owner} )
												file_ids_cooling_off_lock.acquire()
												file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
												file_ids_cooling_off_lock.release()
												found_a_parent = True
											break
											break
					if not found_a_parent:
						# maybe it was trashed by someone and it currently has the service account's "My Drive" as a parent
						was_in_fact_trashed = False
						for parent in data['parents']:
							if parent['isRoot']:
								result = gd.get_file( parent['id'] )
								if result!=False and result['title']=="My Drive":
									# ok! looks like it was trashed
									was_in_fact_trashed = True
									break
						if was_in_fact_trashed:
							sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> file was trashed, setting the trash flag\n" )
							file_ids_cooling_off_lock.acquire()
							file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
							file_ids_cooling_off_lock.release()
							result = gd.trash( file_id )
							file_ids_cooling_off_lock.acquire()
							file_ids_cooling_off[file_id] = int(time.time()) + 10 # expires in 10 seconds
							file_ids_cooling_off_lock.release()
							if result['explicitlyTrashed']==False:
								# so that didn't work
								sys.stdout.write( "* processor " + str(id) + " / " + file_id + " -> I wasn't able to trash the file...\n" )
								mail( error_contact, "unable to trash file", root_owner + " - unable to trash for ID " + file_id )
								keep_going = False

			file_ids_being_processed_lock.acquire()
			file_ids_being_processed.remove( file_id )
			file_ids_being_processed_lock.release()
			processor_threads_status[id] = False
		else:
			# if we didn't find anything we can take it easy a bit
			file_ids_being_processed_lock.release()
			file_ids_to_process_lock.release()
			time.sleep( 0.5 )
	sys.stdout.write( "processor " + str(id) + " finished\n" )


def watcher( gd ):
	global file_ids_cooling_off, file_ids_cooling_off_lock

	sys.stdout.write( "+ watcher starting\n" )

	last_change_id = None
	last_change_id_file = "/var/" + root_folder_id + "-" + root_owner
	sys.stdout.write( last_change_id_file )
	if os.path.isfile( last_change_id_file ):
		f = open( last_change_id_file, "r" )
		last_change_id = f.read()
		f.close()
		sys.stdout.write( "+ starting the watch from " + str(last_change_id) + "\n" )

	while not sigint_received and not fatal_error_reached:
		sys.stdout.write( "+ watch\n" )
		changes = gd.list_changes( last_change_id )
		#pp = pprint.PrettyPrinter(indent=4)
		#pp.pprint( changes )
		for change in changes:
			last_change_id = str(int(change['id']) + 1) # +1 because Google will return the last change ID included so when there is no new change, there is always at least that one
			f = open( last_change_id_file, "w" )
			f.write( last_change_id )
			f.close()
			new_file_id = change['fileId']
			if new_file_id!=root_folder_id and not new_file_id in file_ids_being_processed and not new_file_id in file_ids_to_process: # no point in re-adding a file that is being processed, in fact that's probably why it's popping on the the changes in the first place
				if new_file_id in file_ids_cooling_off.keys():
					if int(time.time())>file_ids_cooling_off[new_file_id]:
						# we're beyond expiration so we'll just clean up the record
						sys.stdout.write( "+ watcher removing expired cool-off for " + new_file_id + "\n" )
						file_ids_cooling_off_lock.acquire()
						del file_ids_cooling_off[new_file_id]
						file_ids_cooling_off_lock.release()
				if new_file_id in file_ids_cooling_off.keys():
					sys.stdout.write( "+ " + new_file_id + " is currently cooling off\n" )
				else:
					sys.stdout.write( "+ adding " + new_file_id + "\n" )
					file_ids_to_process.append( new_file_id )

		time.sleep( 1 )
	sys.stdout.write( "+ watcher finished\n" )


def crawler( id, gd ):
	global crawler_file_stack, crawler_file_stack_fresh, crawler_file_stack_lock, file_ids_to_process, file_ids_to_process_lock, crawler_threads_status, file_ids_cooling_off, file_ids_cooling_off_lock

	sys.stdout.write( "- crawler " + str(id) + " starting\n" )
	while not sigint_received and not fatal_error_reached:
		if id==0 and crawler_file_stack_fresh:
			# the first thread has the special function of bootstrapping a crawl
			sys.stdout.write( "- bootstrapping crawl from root folder ID\n" )
			#crawler_file_stack_lock.acquire() # atomic
			crawler_file_stack.append( root_folder_id )
			#crawler_file_stack_lock.release()
			crawler_file_stack_fresh = False
		elif id==0 and not crawler_file_stack_fresh and len(crawler_file_stack)==0:
			# it also has the function of completing a crawl
			sys.stdout.write( "- we may be done crawling, let's see...\n" )
			# we take a good break after we're done crawling
			sleep_for = 60
			while sleep_for>0:
				if sigint_received or fatal_error_reached:
					return False
				time.sleep( 1 )
				sleep_for = sleep_for - 1
			all_not_working = True
			for crawler_thread_status in crawler_threads_status:
				if crawler_thread_status:
					all_not_working = False
					break
			if all_not_working and len(crawler_file_stack)==0: # nothing new came up we're starting from scratch
				sys.stdout.write( "- yup, we're done crawling, starting over\n" )
				crawler_file_stack_fresh = True # TODO, do we want to wait for file_ids_to_process to be empty before we start throwing more in there?
			else :
				sys.stdout.write( "- false alarm, other threads finished and added more stuff to crawl through\n" )
		else:
			crawler_file_stack_lock.acquire()
			if len(crawler_file_stack)==0:
				crawler_file_stack_lock.release()
				# nothing to do here
				time.sleep( 1 )
			else:
				crawler_threads_status[id] = True # True means not working
				random_index = random.randrange( 0,len(crawler_file_stack) )
				file_id = crawler_file_stack.pop( random_index )
				#print len(crawler_file_stack)
				crawler_file_stack_lock.release()
				new_file_ids = gd.list_children( file_id )
				#crawler_file_stack_lock.acquire() # TODO is append an atomic operation? this may not be necessary this is indeed aotmic http://effbot.org/pyfaq/what-kinds-of-global-value-mutation-are-thread-safe.htm
				#file_ids_to_process_lock.acquire() # TODO is append an atomic operation? this may not be necessary
				for new_file_id in new_file_ids:
					new_file_id = new_file_id['id'].encode("ascii")
					if not new_file_id in file_ids_being_processed and not new_file_id in file_ids_to_process and new_file_id not in all_folder_ids:
						# the crawlers do not add files if the stack has more entries than there are processors to treat them.
						# the watcher is not suject to this limitation, watching always takes priority and will grow the stack however much is necessary
						while not sigint_received and not fatal_error_reached and len(file_ids_to_process)>=processor_threads_count:
							time.sleep( random.random()*10 )
						if new_file_id in file_ids_cooling_off.keys():
							if int(time.time())>file_ids_cooling_off[new_file_id]:
								# we're beyond expiration so we'll just clean up the record
								sys.stdout.write( "- crawler removing expired cool-off for " + new_file_id + "\n" )
								file_ids_cooling_off_lock.acquire()
								del file_ids_cooling_off[new_file_id]
								file_ids_cooling_off_lock.release()
						if new_file_id in file_ids_cooling_off.keys():
							sys.stdout.write( "- " + new_file_id + " is currently cooling off\n" )
						else:
							sys.stdout.write( "- adding " + new_file_id + "\n" )
							file_ids_to_process.append( new_file_id ) # TODO, only add if it's not there already? # TODO make sure the root doesn't get added in there
						crawler_file_stack.append( new_file_id )
					if sigint_received or fatal_error_reached:
						break
				#file_ids_to_process_lock.release()
				#crawler_file_stack_lock.release()
				crawler_threads_status[id] = False # False means not working
	sys.stdout.write( "- crawler " + str(id) + " finished\n" )


class google_drive:
	"class to handle Google Drive interaction"
	access_tokens = {}


	def __init__( self, domain_account_json_file ):
		# loading JSON file config
		domain_account_config_handle = open( domain_account_json_file )
		domain_account_config = json.load( domain_account_config_handle )
		self.client_id = domain_account_config['client_id']
		self.client_email = domain_account_config['client_email']
		self.private_key = domain_account_config['private_key']
		#pp = pprint.PrettyPrinter(indent=4)
		#pp.pprint( self.get_about() )


	def add_parent( self, file_id, parent_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/parents"
		request_type = "POST"
		request_get = ""
		request_post = "{\"id\":\"" + parent_id + "\"}"
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to add_parent\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def delete_parent( self, file_id, parent_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/parent/" + parent_id
		request_type = "DELETE"
		request_get = ""
		request_post = ""
		request_headers = []
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to delete\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def empty_trash( self ):
		request_url = "https://www.googleapis.com/drive/v2/files/trash"
		request_type = "DELETE"
		request_get = ""
		request_post = ""
		request_headers = []
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=204:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + "went away before I ha\n" )
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 204 back from Google trying to delete\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		return True


	def get_parent( self, file_id, parent_id, sub_as=None ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/parents/" + parent_id
		request_type = "GET"
		request_get = ""
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		if sub_as==None:
			sub_as = root_owner
		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, sub_as )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to get_parent\n" )
				return False
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to get_parent\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def list_group_members( self, group_key, page_token=None ):
		request_url = "https://www.googleapis.com/admin/directory/v1/groups/" + group_key + "/members"
		request_type = "GET"
		request_get = ""
		if page_token!=None:
			request_get += "?pageToken=" + urllib.quote_plus( page_token )
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, "need_an_admin_account_that_can_read_groups_here@apps.domain.com" )
		#print result
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to list_group_members\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		if data.has_key( 'nextPageToken' ):
			return data['members'] + self.list_group_members( group_key, data['nextPageToken'] )
		else:
			return data['members']


	def get_file( self, file_id, sub_as=None ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id
		request_type = "GET"
		request_get = ""
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		if sub_as==None:
			sub_as = root_owner
		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, sub_as )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to get_file\n" )
				return False
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to get_file\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def is_file_trashed( self, file_id, data=None ):
		if data!=None:
			request_url = "https://www.googleapis.com/drive/v2/files/" + file_id
			request_type = "GET"
			request_get = ""
			request_post = ""
			request_headers = ["application/x-www-form-urlencoded"]
			request_body = ""

			result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
			if result['response_code']!=200:
				if result['response_code']==404:
					sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to is_file_trashed\n" )
					return True # with this the file shouldn't get any more consideration upstream
				else:
					sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to is_file_trashed\n" )
					global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
					fatal_error_reached = True ;
			try:
				data = json.loads( result['response_body'] )
			except:
				print "THIS IS THE DATA" + result['response_body'] + "|||"
				print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
				print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data['labels']['trashed']


	def is_file_writers_can_share( self, file_id, data=None ):
		if data!=None:
			request_url = "https://www.googleapis.com/drive/v2/files/" + file_id
			request_type = "GET"
			request_get = ""
			request_post = ""
			request_headers = ["application/x-www-form-urlencoded"]
			request_body = ""

			result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
			if result['response_code']!=200:
				if result['response_code']==404:
					sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to is_file_writers_can_share\n" )
					return False # with this the file shouldn't get any more consideration upstream
				else:
					sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to is_file_trashed\n" )
					global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
					fatal_error_reached = True ;
			try:
				data = json.loads( result['response_body'] )
			except:
				print "THIS IS THE DATA" + result['response_body'] + "|||"
				print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
				print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data['writersCanShare']


	def debug_get_about( self ):
		request_url = "https://www.googleapis.com/drive/v2/about"
		request_type = "GET"
		request_get = ""
		request_post = ""
		request_headers = []
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource went away before I had a chance to is_file_trashed\n" )
				return True # with this the file shouldn't get any more consideration upstream
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to is_file_trashed\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def list_changes( self, last_change_id, page_token=None ):
		request_url = "https://www.googleapis.com/drive/v2/changes"
		request_type = "GET"
		request_get = "maxResults=1000&includeDeleted=false&spaces=drive"
		if page_token!=None:
			request_get += "&pageToken=" + urllib.quote_plus( page_token )
		if last_change_id!=None:
			request_get += "&startChangeId=" + last_change_id
		request_post = ""
		request_headers = []
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to list_changes\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		if data.has_key( 'nextPageToken' ):
			return data['items'] + self.list_changes( last_change_id, data['nextPageToken'] )
		else:
			return data['items']


	def list_children( self, folder_id, page_token=None ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + folder_id + "/children"
		request_type = "GET"
		request_get = "maxResults=1000"
		if page_token!=None:
			request_get += "&pageToken=" + urllib.quote_plus( page_token )
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + folder_id + " went away before I had a chance to list_children\n" )
				return [] # as if there were no children
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to list_children\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		if data.has_key( 'nextPageToken' ):
			return data['items'] + self.list_children(folder_id, data['nextPageToken'])
		else:
			return data['items']


	def search_children( self, folder_id, name, page_token=None ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + folder_id + "/children"
		request_type = "GET"
		request_get = "maxResults=1000&q=title%3D'" + urllib.quote_plus( name ) + "'"
		if page_token!=None:
			request_get += "&pageToken=" + urllib.quote_plus( page_token )
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + folder_id + " went away before I had a chance to list_children\n" )
				return [] # as if there were no children
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to list_children\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		if data.has_key( 'nextPageToken' ):
			return data['items'] + self.list_children(folder_id, data['nextPageToken'])
		else:
			return data['items']


	def create_folder( self, parent_folder_id, name ):
		request_url = "https://www.googleapis.com/drive/v2/files"
		request_type = "POST"
		request_get = ""
		request_post = "{\"title\":\"" + name + "\",\"parents\":[{\"id\":\"" + parent_folder_id + "\"}],\"mimeType\": \"application/vnd.google-apps.folder\"}"
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to create folder\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def list_permissions( self, file_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/permissions"
		request_type = "GET"
		request_get = ""
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to list_permissions\n" )
				return [] # as if there were no permissions
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to is_file_trashed\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + str(type(result['response_body']))+ "|||"
		return data['items']


	def list_revisions( self, file_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/revisions"
		request_type = "GET"
		request_get = ""
		request_post = ""
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to list_revisions\n" )
				return [] # as if there were no permissions
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to is_file_trashed\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + str(type(result['response_body']))+ "|||"
		return data['items']


	def change_ownership( self, file_id, current_owner_permission, reference_owner_permission ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/permissions"
		request_type = "POST"
		request_get = "sendNotificationEmails=false"
		request_post = "{\"role\":\"" + reference_owner_permission['role'] + "\",\"type\":\""+ reference_owner_permission['type'] +"\",\"value\":\"" + reference_owner_permission['emailAddress'] + "\"}"
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, current_owner_permission['emailAddress'] )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to change_ownership\n" )
				print result
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to change_ownership\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def delete_permission( self, file_id, permission ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/permissions/" + permission['id']
		request_type = "DELETE"
		request_get = ""
		request_post = ""
		request_headers = []
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=204:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + "went away before I had a chance to delete_permission\n" )
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 204 back from Google trying to delete_permission\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		# no point in doing the following, DELETE returns an empty body
		#try:
		#	data = json.loads( result['response_body'] )
		#except:
		#	print "THIS IS THE DATA" + result['response_body'] + "|||"
		#	print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
		#	print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return True


	def set_writers_can_share_to_false( self, file_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id
		request_type = "PATCH"
		request_get = ""
		request_post = "{\"writersCanShare\":false}"
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource went away before I had a chance to insert_permission\n" )
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to insert_permission\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def trash( self, file_id ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/trash"
		request_type = "POST"
		request_get = ""
		request_post = ""
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to insert_permission\n" )
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to trash\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def insert_permission( self, file_id, permission ):
		request_url = "https://www.googleapis.com/drive/v2/files/" + file_id + "/permissions"
		request_type = "POST"
		request_get = "sendNotificationEmails=false"
		request_post = "{\"role\":\"" + permission['role'] + "\",\"type\":\""+ permission['type'] +"\",\"value\":\"" + permission['emailAddress'] + "\"}"
		request_headers = ["Content-Type: application/json"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, root_owner )
		if result['response_code']!=200:
			if result['response_code']==404:
				sys.stderr.write( "WARNING: it looks like the resource with id " + file_id + " went away before I had a chance to insert_permission\n" )
				return 404
			else:
				sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to insert_permission\n" )
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
		try:
			data = json.loads( result['response_body'] )
		except:
			print "THIS IS THE DATA" + result['response_body'] + "|||"
			print "THIS IS THE DATA" + str(len(result['response_body'])) + "|||"
			print "THIS IS THE DATA" + type(result['response_body'])+ "|||"
		return data


	def refresh_access_token( self, sub_as=None ):
		# building JWT
		jwt_request_date = time.time()
		jwt_expiration_date = jwt_request_date + 60*60 # +1 hour
		jwt_header = '{"alg":"RS256","typ":"JWT"}' ;
		jwt_claim_set = '{"iss":"' + self.client_email + '","scope":"https://www.googleapis.com/auth/drive https://www.googleapis.com/auth/admin.directory.group.member.readonly https://www.googleapis.com/auth/admin.directory.group.member https://www.googleapis.com/auth/admin.directory.group.readonly https://www.googleapis.com/auth/admin.directory.group","aud":"https://www.googleapis.com/oauth2/v3/token","exp":' + str(jwt_expiration_date) + ',"iat":' + str(jwt_request_date) ;
		if sub_as!=None and sub_as!=self.client_email:
			jwt_claim_set += ',"sub":"' + sub_as + '"'
		jwt_claim_set += '}'

		jwt_signature = self.sign_data( self.private_key, base64.urlsafe_b64encode(jwt_header) + '.' + base64.urlsafe_b64encode(jwt_claim_set) )
		jwt = base64.urlsafe_b64encode( jwt_header ) + "." + base64.urlsafe_b64encode( jwt_claim_set ) + "." + base64.urlsafe_b64encode( jwt_signature )

		request_url = "https://www.googleapis.com/oauth2/v3/token"
		request_type = "POST"
		request_get = ""
		request_post = "grant_type=" + urllib.quote_plus( "urn:ietf:params:oauth:grant-type:jwt-bearer" ) + "&assertion=" + jwt
		request_headers = ["application/x-www-form-urlencoded"]
		request_body = ""

		result = self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body )
		if result['response_code']!=200:
			sys.stderr.write( "ERROR: didn't get a 200 back from Google trying to refresh_access_token\n" )
			global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
			fatal_error_reached = True ;
		access_token = json.loads( result['response_body'] )
		if sub_as==None:
			sub_as = self.client_email
		self.access_tokens[sub_as] = {'token':access_token, 'expires_at':jwt_expiration_date}


	def sign_data( self, private_key, data ):
	    '''
	    param: private_key_loc Path to your private key
	    param: package Data to be signed
	    return: base64 encoded signature
	    '''
	    from Crypto.PublicKey import RSA
	    from Crypto.Signature import PKCS1_v1_5
	    from Crypto.Hash import SHA256
	    from base64 import b64encode, b64decode
	    rsakey = RSA.importKey( private_key )
	    signer = PKCS1_v1_5.new( rsakey )
	    digest = SHA256.new()
	    digest.update( data )
	    sign = signer.sign( digest )
	    return sign


	def make_api_call( self, request_url, request_type, request_get, request_post, request_headers, request_body, inject_token_of=None, exponential_backoff_counter=None ):

		# getting an access token if we don't have one already
		if inject_token_of!=None and inject_token_of not in self.access_tokens.keys():
			self.refresh_access_token( inject_token_of )
		# let's try and be proactive about refreshing tokens right before they expire
		if inject_token_of!=None and inject_token_of in self.access_tokens.keys() and (time.time()+10)>self.access_tokens[inject_token_of]['expires_at']:
			self.refresh_access_token( inject_token_of )

		body = StringIO()
		headers = StringIO()
		c = pycurl.Curl()
		c.setopt( c.CONNECTTIMEOUT, 5 )
		c.setopt( c.TIMEOUT, 5 )
		c.setopt( c.NOSIGNAL, 5 )
		if request_get==None or request_get=="":
			c.setopt( c.URL, request_url.encode("ascii") )
		else:
			c.setopt( c.URL, request_url.encode("ascii") + "?" + request_get.encode("ascii") )
		if not( request_post==None or request_post=="" ):
			c.setopt( c.POSTFIELDS, request_post.encode("ascii") )
			c.setopt( c.POST, 1 )
		c.setopt( c.FOLLOWLOCATION, True )
		c.setopt( c.CUSTOMREQUEST, request_type )
		add_to_headers = ["Content-length: " +  str(len(request_post))]
		if inject_token_of!=None:
			add_to_headers.append( "Authorization: Bearer " + self.access_tokens[inject_token_of]['token']['access_token'].encode("ascii") )
		c.setopt( c.HTTPHEADER, request_headers + add_to_headers )
		c.setopt( c.WRITEFUNCTION, body.write )
		c.setopt( c.HEADERFUNCTION, headers.write )
		try:
			c.perform()
		except:
			pass
		response_code = c.getinfo( c.RESPONSE_CODE )
		response_body = body.getvalue()
		response_headers = headers.getvalue()
		c.close()
		body.close()
		headers.close()

		if response_code==200 or response_code==204 or response_code==404: # 200 or 204 when everything is cool; 404 when the resource went away before we had a chance to do anything with it. We pass that back so the originating functions get a chances to handle it properly.
			return {'response_code':response_code, 'response_body':response_body, 'response_headers':response_headers}
		elif response_code==401:
			# todo, we need a new access token here
			if inject_token_of==None:
				sys.stderr.write( "ERROR: I've gotten a 401 back from Google when I wasn't even supposed to be injecting a token; something's gone horribly wrong.\n" ) ;
				global fatal_error_reached # if I don't put it everywhere I have an error I'll forget and it won't kick in, nice job Python.
				fatal_error_reached = True ;
			else:
				sys.stderr.write( "WARNING: the logic should be set so that we shouldn't get 401s back from Google\n" )
				self.refresh_access_token( inject_token_of )
				# and then we try again
				return self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, inject_token_of, exponential_backoff_counter )
		else:
			sys.stdout.write( "exponential backoff with http response code " + str(response_code) + " url:" + request_url + " body:" + response_body + "body size:" + str(len(response_body)) + "\n" )
			# in any other case, we implement exponential backoff (https://developers.google.com/drive/web/handle-errors)
			if exponential_backoff_counter==None:
				exponential_backoff_counter = 0
			exponential_backoff_counter += 1
			# Google say to exponential back-off up to 5 but I've seen enough weird things that we do up to ~2 hours
			if exponential_backoff_counter>12:
				sys.stdout.write( "exponential backoff limit exceeded, gotta return...\n" )
				# well, we've tried and tried to no avail, time to return what we are getting
				return {'response_code':response_code, 'response_body':response_body, 'response_headers':response_headers}
			else:
				# TODO, if we've reached a global error state or SIGINT we actually don't want to bother with the exponential backoff
				sleep_for = math.pow( 2, exponential_backoff_counter )
				#print "sleeping for " +  str(sleep_for)
				while sleep_for>0:
					if sigint_received or fatal_error_reached:
						return False ;
					time.sleep( 1 )
					sleep_for = sleep_for - 1
				return self.make_api_call( request_url, request_type, request_get, request_post, request_headers, request_body, inject_token_of, exponential_backoff_counter )



# main
gd = google_drive( domain_delegated_account_json_file )
reference_permissions = gd.list_permissions( root_folder_id )

# we need to know which other folders are being considered by other forks of the program in case one is a child of the one we're working on
all_folder_ids = open( "config.txt" )
all_folder_ids = all_folder_ids.read()
all_folder_ids = all_folder_ids.split( "\n" )
all_folder_ids = filter( None, all_folder_ids )
for i in range( len(all_folder_ids) ):
	all_folder_ids[i] = all_folder_ids[i].split( " " )
	all_folder_ids[i] = all_folder_ids[i][0]


for i in range( crawler_threads_count ):
	crawler_threads_status.append( False ) # False means not working
  	t = threading.Thread( target=crawler, args=(i, gd) )
  	t.start()

for i in range( processor_threads_count ):
	processor_threads_status.append( False ) # False means not working
   	t = threading.Thread( target=processor, args=(i, gd, reference_permissions) )
   	t.start()

# no point in having more than one watcher
t = threading.Thread( target=watcher, args=(gd,) ) # I hate Python
t.start()

while True:
	threads = threading.enumerate()
	if len(threads) == 1:
		break
	for t in threads:
		if t != threading.currentThread():
			t.join(10)
